#include <Servo.h>

// Sensors
#define IR_0 A5 
#define IR_1 A4 
#define Echo A2
#define Trig A3

// Motor Driver
#define enA 5
#define in1 6
#define in2 7
#define in3 8
#define in4 9
#define enB 10

// Servo
#define servoPin 11
Servo myservo;

// LED
#define LED 12

double motorA = 120;
double motorB = 120;

int thres = 6;

unsigned long previousMillis = 0;
unsigned long interval = 100; // General loop interval

// Obstacle handling state machine
int avoidStage = 0;
unsigned long avoidTimer = 0;
bool avoiding = false;

void setup() {
  Serial.begin(9600);

  pinMode(IR_0, INPUT);
  pinMode(IR_1, INPUT);
  pinMode(Echo, INPUT);
  pinMode(Trig, OUTPUT);
  pinMode(LED, OUTPUT);

  pinMode(enA, OUTPUT);
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  pinMode(in3, OUTPUT);
  pinMode(in4, OUTPUT);
  pinMode(enB, OUTPUT);

  analogWrite(enA, motorA);
  analogWrite(enB, motorB);

  myservo.attach(servoPin);
  //digitalWrite(LED, HIGH);

  myservo.write(0);
}

void loop() {
  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis < interval) return;
  previousMillis = currentMillis;

  if (avoiding) {
    handleObstacle();
    return;
  }

  int distance_U = Ultrasonic_Read();
  int res_IR = digitalRead(IR_0);
  int res_Bk = digitalRead(IR_1);

  Serial.print("Ultrasonic: ");
  Serial.print(distance_U);
  Serial.print(" cm | IR: ");
  Serial.println(res_IR);
  Serial.println(res_Bk);

  if (res_IR == 0) {
    Serial.print(" EDGE! ");
    stopMotors();
    backward(motorA, motorB);
    delay(700);
    stopMotors();
    avoidTimer = millis();
    avoidStage = 10; // Use custom state for IR edge
    avoiding = true;
    return;
  }

    if (res_Bk == 0) {
    Serial.print(" EDGE BACK! ");
    stopMotors();
    forward(motorA, motorB);
    delay(700);
    stopMotors();
    avoidTimer = millis();
    avoidStage = 10; // Use custom state for IR edge
    avoiding = true;
    return;
  }

  if (distance_U > 0 && distance_U < thres) {
    Serial.print(" hit! ");
    stopMotors();
    //Blink();
    avoidStage = 0;
    avoidTimer = millis();
    avoiding = true;
    return;
  }

  if (distance_U > 60 || distance_U == -1) {
    Serial.println("No object detected, rotating...");
    rotateAngle(180, true);
    return;
  }

  if (distance_U > thres && distance_U <= 60) {
    Serial.println("Object detected, moving forward...");
    forward(motorA, motorB);
    return;
  }

  stopMotors();
}

void handleObstacle() {
  unsigned long now = millis();

  switch (avoidStage) {
    case 0:
      sweepServo();
      avoidTimer = now;
      avoidStage = 1;
      break;

    case 1:
      if (now - avoidTimer > 500) {
        //digitalWrite(LED, LOW);
        rotateAngle(180, false);
        avoiding = false;
      }
      break;

    case 10:
      if (now - avoidTimer > 500) {
        rotateAngle(180, true);
        avoiding = false;
      }
      break;
  }
}

long Ultrasonic_Read() {
  digitalWrite(Trig, LOW);
  delayMicroseconds(2);
  digitalWrite(Trig, HIGH);
  delayMicroseconds(10);
  long time = pulseIn(Echo, HIGH, 30000);
  if (time == 0) return -1;
  return time / 29 / 2;
}

void Blink() {
  digitalWrite(LED, HIGH);
  Serial.print(" led on ");
}

void sweepServo() {
  Blink();
  Serial.println("sweeping");
  for (int angle = 0; angle <= 2; angle += 2) {
    digitalWrite(LED, HIGH);
    myservo.write(angle);
    delay(50);
    digitalWrite(LED, LOW);
  }
    for (int angle =2; angle <= 4; angle += 2) {
    digitalWrite(LED, HIGH);
    myservo.write(angle);
    delay(50);
    digitalWrite(LED, LOW);
  }
  for (int angle = 4; angle <= 170; angle += 2) {
    digitalWrite(LED, HIGH);
    myservo.write(angle);
    delay(50);
    digitalWrite(LED, LOW);
  }
    for (int angle = 6; angle <= 170; angle += 10) {
    digitalWrite(LED, HIGH);
    myservo.write(angle);
    delay(500);
    digitalWrite(LED, LOW);
  }
  delay(100);
  for (int angle = 170; angle >= 10; angle -= 90) {
    digitalWrite(LED, HIGH);
    myservo.write(angle);
    delay(300);
    digitalWrite(LED, LOW);
  }
  delay(100);
  myservo.write(20);
    backward(motorA, motorB);
    delay(500);
    stopMotors();
  digitalWrite(LED, LOW);
}

void forward(double motorA, double motorB) {
  analogWrite(enA, motorA);
  analogWrite(enB, motorB);

  Serial.print(" forward  ");
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
  digitalWrite(in3, LOW);
  digitalWrite(in4, HIGH);
}

void turnRight(){
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
  digitalWrite(in3, HIGH);
  digitalWrite(in4, LOW);
}

void turnLeft(){
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  digitalWrite(in3, LOW);
  digitalWrite(in4, HIGH);
}

void backward(double motorA, double motorB){
  Serial.println("back");
  analogWrite(enA, motorA);
  analogWrite(enB, motorB);

  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  digitalWrite(in3, HIGH);
  digitalWrite(in4, LOW);
}

void rotate(double motorA, double motorB){
  analogWrite(enA, motorA);
  analogWrite(enB, motorB);

  int bestAngle = 90;
  long minDist = 1000;

  for (int angle = 50; angle <= 130; angle += 20) {
    turnLeft();
    delay(100);
    long d = Ultrasonic_Read();
    if (d > 0 && d < minDist) {
      minDist = d;
      bestAngle = angle;
    }
  }
  if (bestAngle < 85) {
    Serial.println("Turning left towards detected object");
    turnLeft();
  } else if (bestAngle > 95) {
    Serial.println("Turning right towards detected object");
    turnRight();
  } else {
    Serial.println("Forward direction");
    return;
  }

  delay(500);  
  stopMotors();
}

void rotateAngle(int angleDegrees, bool clockwise) {
  double rotationSpeed = 120;  // adjust to your robot
  int durationPerDegree = 10;  // ms per degree (you need to calibrate)

  analogWrite(enA, rotationSpeed);
  analogWrite(enB, rotationSpeed);

  if (clockwise) {
    turnLeft();
  } else {
    turnRight();
  }

  delay(angleDegrees * durationPerDegree);
  stopMotors();
}

void stopMotors() {
  digitalWrite(in1, LOW);
  digitalWrite(in2, LOW);
  digitalWrite(in3, LOW);
  digitalWrite(in4, LOW);

  delay(100);
}